﻿// COPYRIGHT 2014, 2015 by the Open Rails project.
// 
// This file is part of Open Rails.
// 
// Open Rails is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Open Rails is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Open Rails.  If not, see <http://www.gnu.org/licenses/>.

using Ionic.Zip;
using Newtonsoft.Json;
using ORTS.Settings;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net;
using System.Runtime.InteropServices;
using System.Security.Cryptography;
using System.Security.Cryptography.Pkcs;
using System.Security.Cryptography.X509Certificates;
using System.Security.Principal;
using System.Text;
using System.Threading;

namespace ORTS.Updater
{
    public class UpdateManager
    {
        // The date on this is fairly arbitrary - it's only used in a calculation to round the DateTime up to the next TimeSpan period.
        readonly DateTime BaseDateTimeMidnightLocal = new DateTime(2010, 1, 1, 0, 0, 0, DateTimeKind.Local);

        public const string ChannelCommandLine = "/CHANNEL=";
        public const string WaitProcessIdCommandLine = "/WAITPID=";
        public const string RelaunchCommandLine = "/RELAUNCH=";
        public const string ElevationCommandLine = "/ELEVATE=";

        const string TemporaryDirectoryName = "Open Rails Updater Temporary Files";

        public event EventHandler<ProgressChangedEventArgs> ApplyProgressChanged;

        readonly string BasePath;
        readonly string ProductName;
        readonly string ProductVersion;
        readonly UpdateSettings Settings;
        readonly UpdateState State;
        UpdateSettings Channel;
        bool Force;

        public string ChannelName { get; set; }
        public string ChangeLogLink { get { return Channel != null ? Channel.ChangeLogLink : null; } }
        public Update LastUpdate { get; private set; }
        public Exception LastCheckError { get; private set; }
        public Exception LastUpdateError { get; private set; }
        public bool UpdaterNeedsElevation { get; private set; }

        public static string GetMainExecutable(string pathName, string productName)
        {
            foreach (var file in Directory.GetFiles(pathName, "*.exe"))
            {
                try
                {
                    var versionInfo = FileVersionInfo.GetVersionInfo(file);
                    if (versionInfo.FileDescription == productName)
                        return file;
                }
                catch { }
            }
            return null;
        }

        public UpdateManager(string basePath, string productName, string productVersion)
        {
            if (!Directory.Exists(basePath)) throw new ArgumentException("The specified path must be valid and exist as a directory.", "basePath");
            BasePath = basePath;
            ProductName = productName;
            ProductVersion = productVersion;
            try
            {
                Settings = new UpdateSettings();
                State = new UpdateState();
                Channel = new UpdateSettings(ChannelName = Settings.Channel);
            }
            catch (ArgumentException)
            {
                // Updater.ini doesn't exist. That's cool, we'll just disable updating.
            }

            // Check for elevation to update; elevation is needed if the update writes failed and the user is NOT an
            // Administrator. Weird cases (like no permissions on the directory for anyone) are not handled.
            try
            {
                TestUpdateWrites();
            }
            catch
            {
                var identity = WindowsIdentity.GetCurrent();
                var principal = new WindowsPrincipal(identity);
                UpdaterNeedsElevation = !principal.IsInRole(WindowsBuiltInRole.Administrator);
            }
        }

        public string[] GetChannels()
        {
            if (Channel == null)
                return new string[0];

            return Settings.GetChannels();
        }

        public void SetChannel(string channel)
        {
            if (Channel == null)
                throw new InvalidOperationException();

            // Switch channel and save the change.
            Settings.Channel = channel;
            Settings.Save();
            Channel = new UpdateSettings(ChannelName = Settings.Channel);

            // Do a forced update check because the cached update data is likely to only be valid for the old channel.
            Force = true;
        }

        public void Check()
        {
            // If there's no updater file or the update channel is not correctly configured, exit without error.
            if (Channel == null)
                return;

            try
            {
                // If we're not at the appropriate time for the next check (and we're not forced), we reconstruct the cached update/error and exit.
                if (DateTime.Now < State.NextCheck && !Force)
                {
                    LastUpdate = State.Update.Length > 0 ? JsonConvert.DeserializeObject<Update>(State.Update) : null;
                    LastCheckError = State.Update.Length > 0 || string.IsNullOrEmpty(Channel.URL) ? null : new InvalidDataException("Last update check failed.");

                    // Validate that the deserialized update is sane.
                    ValidateLastUpdate();

                    return;
                }

                // This updates the NextCheck time and clears the cached update/error.
                ResetCachedUpdate();

                if (string.IsNullOrEmpty(Channel.URL))
                {
                    // If there's no update URL, reset cached update/error.
                    LastUpdate = null;
                    LastCheckError = null;
                    return;
                }

                // Fetch the update URL (adding ?force=true if forced) and cache the update/error.
                var client = new WebClient()
                {
                    CachePolicy = new System.Net.Cache.RequestCachePolicy(System.Net.Cache.RequestCacheLevel.BypassCache),
                    Encoding = Encoding.UTF8,
                };
                client.Headers[HttpRequestHeader.UserAgent] = GetUserAgent();
                var updateUri = new Uri(!Force ? Channel.URL : Channel.URL.Contains('?') ? Channel.URL + "&force=true" : Channel.URL + "?force=true");
                var updateData = client.DownloadString(updateUri);
                LastUpdate = JsonConvert.DeserializeObject<Update>(updateData);
                LastCheckError = null;

                // Check it's all good.
                ValidateLastUpdate();

                CacheUpdate(updateData);
            }
            catch (Exception error)
            {
                // This could be a problem deserializing the LastUpdate or fetching/deserializing the new update. It doesn't really matter, we record an error.
                LastUpdate = null;
                LastCheckError = error;
                Trace.WriteLine(error);

                ResetCachedUpdate();
            }
        }

        void ValidateLastUpdate()
        {
            if (LastUpdate != null)
            {
                var uri = new Uri(LastUpdate.Url, UriKind.RelativeOrAbsolute);
                if (uri.IsAbsoluteUri)
                {
                    LastUpdate = null;
                    LastCheckError = new InvalidDataException("Update URL must be relative to channel URL.");
                }
            }
        }

        public void Update()
        {
            if (LastUpdate == null) throw new InvalidOperationException("Cannot get update when no LatestUpdate exists.");
            try
            {
                var process = Process.Start(FileUpdater, String.Format("{0}{1} {2}{3} {4}{5} {6}{7}", ChannelCommandLine, ChannelName, WaitProcessIdCommandLine, Process.GetCurrentProcess().Id, RelaunchCommandLine, 1, ElevationCommandLine, UpdaterNeedsElevation ? 1 : 0));
                process.WaitForInputIdle();
                Environment.Exit(0);
            }
            catch (Exception error)
            {
                LastUpdateError = error;
                return;
            }
        }

        public void Apply()
        {
            if (LastUpdate == null) throw new InvalidOperationException("There is no update to apply.");

            TriggerApplyProgressChanged(0);
            try
            {
                TestUpdateWrites();
                TriggerApplyProgressChanged(1);

                CleanDirectories();
                TriggerApplyProgressChanged(2);

                DownloadUpdate(2, 65);
                TriggerApplyProgressChanged(67);

                ExtractUpdate(67, 30);
                TriggerApplyProgressChanged(97);

                if (UpdateIsReady())
                {
                    TriggerApplyProgressChanged(98);

                    VerifyUpdate();
                    TriggerApplyProgressChanged(99);

                    ApplyUpdate();
                    TriggerApplyProgressChanged(100);
                }

                LastUpdateError = null;
            }
            catch (Exception error)
            {
                LastUpdateError = error;
            }
            finally
            {
                CleanDirectories();
            }
        }

        void TriggerApplyProgressChanged(int progressPercentage)
        {
            var progressEvent = ApplyProgressChanged;
            if (progressEvent != null)
                progressEvent(this, new ProgressChangedEventArgs(progressPercentage, null));
        }

        string GetUserAgent()
        {
            return String.Format("{0}/{1}", ProductName, ProductVersion);
        }

        void ResetCachedUpdate()
        {
            State.LastCheck = DateTime.Now;
            // So what we're doing here is rounding up the DateTime (LastCheck) to the next TimeSpan (TTL) period. For
            // example, if the TTL was 1 hour, we'd round up the the start of the next hour. Similarly, if the TTL was
            // 1 day, we'd round up to midnight (the start of the next day). The purpose of this is to avoid 2 * TTL 
            // checking which might well occur if you always launch Open Rails around the same time of day each day -
            // if they launch it at 6:00PM on Monday, then 5:30PM on Tuesday, they won't get an update chech on
            // Tuesday. With the time rounding, they should get one check/day if the TTL is 1 day and they open it
            // every day. (This is why BaseDateTimeMidnightLocal uses the local midnight!)
            State.NextCheck = Channel.TTL.TotalMinutes > 1 ? BaseDateTimeMidnightLocal.AddSeconds(Math.Ceiling((State.LastCheck - BaseDateTimeMidnightLocal).TotalSeconds / Channel.TTL.TotalSeconds) * Channel.TTL.TotalSeconds) : State.LastCheck + TimeSpan.FromMinutes(1);
            State.Update = "";
            State.Save();
        }

        void CacheUpdate(string updateData)
        {
            Force = false;
            State.Update = updateData;
            State.Save();
        }

        string PathUpdateTest { get { return Path.Combine(BasePath, "UpdateTest"); } }
        string PathUpdateDirty { get { return Path.Combine(BasePath, "UpdateDirty"); } }
        string PathUpdateStage { get { return Path.Combine(BasePath, "UpdateStage"); } }
        string PathDocumentation { get { return Path.Combine(BasePath, "Documentation"); } }
        string PathUpdateDocumentation { get { return Path.Combine(PathUpdateStage, "Documentation"); } }
        string FileUpdateStage { get { return Path.Combine(PathUpdateStage, "Update.zip"); } }
        string FileUpdateStageIsReady { get { return GetMainExecutable(PathUpdateStage, ProductName); } }
        string FileSettings { get { return Path.Combine(BasePath, "OpenRails.ini"); } }
        string FileUpdater { get { return Path.Combine(BasePath, "Updater.exe"); } }

        void TestUpdateWrites()
        {
            Directory.CreateDirectory(PathUpdateTest);
            Directory.Delete(PathUpdateTest, true);
        }

        void CleanDirectories()
        {
            if (Directory.Exists(PathUpdateDirty))
                CleanDirectory(PathUpdateDirty);

            if (Directory.Exists(PathUpdateStage))
                CleanDirectory(PathUpdateStage);
        }

        void CleanDirectory(string path)
        {
            // Clean up as much as we can here, but any in-use files will fail. Don't worry about them. This is
            // called before the update begins so we'll always start from a clean slate.

            // Scan the files in any order.
            var files = Directory.GetFiles(path, "*", SearchOption.AllDirectories);
            foreach (var file in files)
            {
                try { File.Delete(file); }
                catch { }
            }

            // Scan the directories by descending length, so that we never try and delete a parent before a child.
            var directories = Directory.GetDirectories(path, "*", SearchOption.AllDirectories).OrderByDescending(s => s.Length);
            foreach (var directory in directories)
            {
                try { Directory.Delete(directory); }
                catch { }
            }
            try { Directory.Delete(path); }
            catch { }
        }

        void DownloadUpdate(int progressMin, int progressLength)
        {
            if (!Directory.Exists(PathUpdateStage))
                Directory.CreateDirectory(PathUpdateStage);

            var updateUri = new Uri(Channel.URL);
            var uri = new Uri(updateUri, LastUpdate.Url);
            var client = new WebClient();
            AsyncCompletedEventArgs done = null;
            client.DownloadProgressChanged += (object sender, DownloadProgressChangedEventArgs e) =>
            {
                try
                {
                    TriggerApplyProgressChanged(progressMin + progressLength * e.ProgressPercentage / 100);
                }
                catch (Exception error)
                {
                    done = new AsyncCompletedEventArgs(error, false, e.UserState);
                }
            };
            client.DownloadFileCompleted += (object sender, AsyncCompletedEventArgs e) =>
            {
                done = e;
            };
            client.Headers[HttpRequestHeader.UserAgent] = GetUserAgent();

            client.DownloadFileAsync(uri, FileUpdateStage);
            while (done == null)
            {
                Thread.Sleep(100);
            }
            if (done.Error != null)
                throw done.Error;

            TriggerApplyProgressChanged(progressMin + progressLength);
        }

        void ExtractUpdate(int progressMin, int progressLength)
        {
            using (var zip = ZipFile.Read(FileUpdateStage))
            {
                zip.ExtractProgress += (object sender, ExtractProgressEventArgs e) =>
                {
                    if (e.EventType == ZipProgressEventType.Extracting_BeforeExtractEntry)
                        TriggerApplyProgressChanged(progressMin + progressLength * e.EntriesExtracted / e.EntriesTotal);
                };
                zip.ExtractAll(PathUpdateStage, ExtractExistingFileAction.OverwriteSilently);
            }

            File.Delete(FileUpdateStage);

            TriggerApplyProgressChanged(progressMin + progressLength);
        }

        bool UpdateIsReady()
        {
            // The staging directory must exist, contain OpenRails.exe (be ready) and NOT contain the update zip.
            return Directory.Exists(PathUpdateStage)
                && File.Exists(FileUpdateStageIsReady)
                && !File.Exists(FileUpdateStage);
        }

        void VerifyUpdate()
        {
            var files = Directory.GetFiles(PathUpdateStage, "*", SearchOption.AllDirectories);

            var expectedSubjects = new HashSet<string>();
            try
            {
                foreach (var cert in GetCertificatesFromFile(FileUpdater))
                    expectedSubjects.Add(cert.Subject);
            }
            catch (CryptographicException)
            {
                // No signature on the updater, so we can't verify the update. :(
                return;
            }

            for (var i = 0; i < files.Length; i++)
            {
                if (files[i].EndsWith(".cpl", StringComparison.OrdinalIgnoreCase) ||
                    files[i].EndsWith(".dll", StringComparison.OrdinalIgnoreCase) ||
                    files[i].EndsWith(".exe", StringComparison.OrdinalIgnoreCase) ||
                    files[i].EndsWith(".ocx", StringComparison.OrdinalIgnoreCase) ||
                    files[i].EndsWith(".sys", StringComparison.OrdinalIgnoreCase))
                {
                    var certificates = GetCertificatesFromFile(files[i]);
                    if (!certificates.Any(c => expectedSubjects.Contains(c.Subject)))
                        throw new InvalidDataException("Cryptographic signatures don't match. Expected a common subject in old subjects:\n\n" + FormatCertificateSubjectList(expectedSubjects) + "\n\nAnd new subjects:\n\n" + FormatCertificateSubjectList(certificates) + "\n");
                }
            }

            return;
        }

        void ApplyUpdate()
        {
            var basePathFiles = Directory.GetFiles(BasePath, "*", SearchOption.AllDirectories).Where(file =>
                !file.StartsWith(PathUpdateDirty, StringComparison.OrdinalIgnoreCase) &&
                !file.StartsWith(PathUpdateStage, StringComparison.OrdinalIgnoreCase) &&
                // Skip deleting the Documentation directory unless a new one is included in the update.
                (!file.StartsWith(PathDocumentation, StringComparison.OrdinalIgnoreCase) || Directory.Exists(PathUpdateDocumentation)) &&
                !file.Equals(FileSettings, StringComparison.OrdinalIgnoreCase)).ToArray();
            var updateStageFiles = Directory.GetFiles(PathUpdateStage, "*", SearchOption.AllDirectories);

            // Move (almost) all the files from the base path to the dirty path - this removes all the old program files.
            MoveDirectoryFiles(BasePath, PathUpdateDirty, basePathFiles);

            // Move all the files from the stage path to the base path - this adds all the new program files.
            MoveDirectoryFiles(PathUpdateStage, BasePath, updateStageFiles);
        }

        void MoveDirectoryFiles(string source, string destination, string[] files)
        {
            CreateDirectoryLayout(source, destination);

            foreach (var file in files)
                File.Move(file, Path.Combine(destination, GetRelativePath(file, source)));

            foreach (var file in files)
            {
                var directory = Path.GetDirectoryName(file);
                if (!directory.Equals(source, StringComparison.OrdinalIgnoreCase) && Directory.Exists(directory))
                    Directory.Delete(directory);
            }
        }

        void CreateDirectoryLayout(string source, string destination)
        {
            Debug.Assert(Directory.Exists(source));
            var directories = Directory.GetDirectories(source, "*", SearchOption.AllDirectories);
            // Make each directory to its relative form and create them as needed.
            if (!Directory.Exists(destination))
                Directory.CreateDirectory(destination);
            foreach (var directory in from directory in directories select Path.Combine(destination, GetRelativePath(directory, source)))
                if (!Directory.Exists(directory))
                    Directory.CreateDirectory(directory);
        }

        string GetRelativePath(string directory, string basePath)
        {
            Debug.Assert(Path.IsPathRooted(directory));
            Debug.Assert(Path.IsPathRooted(basePath));
            Debug.Assert(Path.GetPathRoot(directory) == Path.GetPathRoot(basePath));
            Debug.Assert(directory.Length > basePath.Length);
            Debug.Assert(directory[basePath.Length] == Path.DirectorySeparatorChar);
            return directory.Substring(basePath.Length + 1);
        }

        static string FormatCertificateSubjectList(IEnumerable<string> subjects)
        {
            return string.Join("\n", subjects.Select(s => "- " + s).ToArray());
        }

        static string FormatCertificateSubjectList(IEnumerable<X509Certificate2> certificates)
        {
            return FormatCertificateSubjectList(certificates.Select(c => c.Subject));
        }

        static List<X509Certificate2> GetCertificatesFromFile(string filename)
        {
            IntPtr cryptMsg = IntPtr.Zero;
            if (!NativeMethods.CryptQueryObject(NativeMethods.CERT_QUERY_OBJECT_FILE, filename, NativeMethods.CERT_QUERY_CONTENT_FLAG_PKCS7_SIGNED_EMBED, NativeMethods.CERT_QUERY_FORMAT_FLAG_ALL, 0, 0, 0, 0, 0, ref cryptMsg, 0))
                throw new Win32Exception(Marshal.GetLastWin32Error());

            // Get size of the encoded message.
            int dataSize = 0;
            if (!NativeMethods.CryptMsgGetParam(cryptMsg, NativeMethods.CMSG_ENCODED_MESSAGE, 0, IntPtr.Zero, ref dataSize))
                throw new Win32Exception(Marshal.GetLastWin32Error());

            // Get the encoded message.
            var data = new byte[dataSize];
            if (!NativeMethods.CryptMsgGetParam(cryptMsg, NativeMethods.CMSG_ENCODED_MESSAGE, 0, data, ref dataSize))
                throw new Win32Exception(Marshal.GetLastWin32Error());

            return GetCertificatesFromEncodedData(data);
        }

        static List<X509Certificate2> GetCertificatesFromEncodedData(byte[] data)
        {
            var certs = new List<X509Certificate2>();

            var signedCms = new SignedCms();
            signedCms.Decode(data);

            foreach (var signerInfo in signedCms.SignerInfos)
            {
                // Record this signer info's certificate if it has one.
                if (signerInfo.Certificate != null)
                    certs.Add(signerInfo.Certificate);

                foreach (var unsignedAttribute in signerInfo.UnsignedAttributes)
                {
                    // This attribute Oid is for "code signatures" and is used to attach multiple signatures to a single item.
                    if (unsignedAttribute.Oid.Value == "1.3.6.1.4.1.311.2.4.1")
                    {
                        foreach (var value in unsignedAttribute.Values)
                            certs.AddRange(GetCertificatesFromEncodedData(value.RawData));
                    }
                }
            }

            return certs;
        }
    }

    public class Update
    {
        [JsonProperty]
        public DateTime Date { get; private set; }

        [JsonProperty]
        public string Url { get; private set; }

        [JsonProperty]
        public string Version { get; private set; }
    }
}
