﻿// COPYRIGHT 2009, 2010, 2011, 2012, 2013, 2014 by the Open Rails project.
// 
// This file is part of Open Rails.
// 
// Open Rails is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Open Rails is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Open Rails.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using Microsoft.Xna.Framework;
using Orts.Parsers.Msts;
using ORTS.Common;

namespace Orts.Formats.Msts
{
    public class WorldFile
    {
        public readonly int TileX, TileZ;
        public readonly Tr_Worldfile Tr_Worldfile;

        public WorldFile(string filename)
            : this(filename, null)
        {
        }

        public WorldFile(string filename, List<TokenID> allowedTokens)
        {
            // Parse the tile location out of the filename.
            var p = filename.ToUpper().LastIndexOf("\\WORLD\\W");
            TileX = int.Parse(filename.Substring(p + 8, 7));
            TileZ = int.Parse(filename.Substring(p + 15, 7));

            using (var sbr = SBR.Open(filename))
            {
                using (var block = sbr.ReadSubBlock())
                {
                    Tr_Worldfile = new Tr_Worldfile(block, filename, allowedTokens);
                }
                // some w files have additional comments at the end 
                //       eg _Skip ( "TS DB-Utility - Version: 3.4.05(13.10.2009), Filetype='World', Copyright (C) 2003-2009 by ...CarlosHR..." )
                sbr.Skip();
            }
        }
    }

    public class Tr_Worldfile : List<WorldObject>
    {
        static HashSet<TokenID> UnknownBlockIDs = new HashSet<TokenID>()
        {
            TokenID.VDbIdCount,
            TokenID.ViewDbSphere,
        };

        public Tr_Worldfile(SBR block, string filename, List<TokenID> allowedTokens)
        {
            block.VerifyID(TokenID.Tr_Worldfile);
            var currentWatermark = 0;
            while (!block.EndOfBlock())
            {
                using (var subBlock = block.ReadSubBlock())
                {
                    try
                    {
                        if (allowedTokens == null || allowedTokens.Contains(subBlock.ID))
                            LoadObject(subBlock, ref currentWatermark, filename);
                        else
                            subBlock.Skip();
                    }
                    catch (Exception error)
                    {
                        Trace.WriteLine(new FileLoadException(filename, error));
                    }
                }
            }
        }

        void LoadObject(SBR subBlock, ref int currentWatermark, string filename)
        {
            switch (subBlock.ID)
            {
                case TokenID.CollideObject:
                case TokenID.Static:
                    Add(new StaticObj(subBlock, currentWatermark));
                    break;
                case TokenID.TrackObj:
                    Add(new TrackObj(subBlock, currentWatermark));
                    break;
                case TokenID.CarSpawner:
                case (TokenID)357:
                    Add(new CarSpawnerObj(subBlock, currentWatermark));
                    break;
                case TokenID.Siding:
                case (TokenID)361:
                    Add(new SidingObj(subBlock, currentWatermark));
                    break;
                case TokenID.Platform:
                    Add(new PlatformObj(subBlock, currentWatermark));
                    break;
                case TokenID.Forest:
                    Add(new ForestObj(subBlock, currentWatermark));
                    break;
                case TokenID.LevelCr:
                    Add(new LevelCrossingObj(subBlock, currentWatermark));
                    break;
                case TokenID.Dyntrack:
                case (TokenID)306:
                    Add(new DyntrackObj(subBlock, currentWatermark));
                    break;
                case TokenID.Transfer:
                case (TokenID)363:
                    Add(new TransferObj(subBlock, currentWatermark));
                    break;
                case TokenID.Gantry:
                case (TokenID)356:
                    // TODO: Add real handling for gantry objects.
                    Add(new BaseObj(subBlock, currentWatermark));
                    break;
                case TokenID.Pickup:
                case (TokenID)359:
                    Add(new PickupObj(subBlock, currentWatermark));
                    break;
                case TokenID.Hazard:
                    //case (TokenID)359:
                    Add(new HazardObj(subBlock, currentWatermark));
                    break;
                case TokenID.Signal:
                    Add(new SignalObj(subBlock, currentWatermark));
                    break;
                case TokenID.Speedpost:
                    Add(new SpeedPostObj(subBlock, currentWatermark));
                    break;
                case TokenID.Tr_Watermark:
                    currentWatermark = subBlock.ReadInt();
                    break;
                default:
                    if (!UnknownBlockIDs.Contains(subBlock.ID))
                    {
                        UnknownBlockIDs.Add(subBlock.ID);
                        Trace.TraceWarning("Skipped unknown world block {0} (0x{0:X}) first seen in {1}", subBlock.ID, filename);
                    }
                    subBlock.Skip();
                    break;
            }
        }
    }

    public class BaseObj : WorldObject
    {
        public BaseObj()
        {
        }

        public BaseObj(SBR block, int detailLevel)
        {
            StaticDetailLevel = detailLevel;

            while (!block.EndOfBlock())
            {
                using (var subBlock = block.ReadSubBlock())
                {
                    switch (subBlock.ID)
                    {
                        case TokenID.UiD: UID = subBlock.ReadUInt(); break;
                        case TokenID.FileName: FileName = subBlock.ReadString(); break;
                        case TokenID.Position: Position = new STFPositionItem(subBlock); break;
                        case TokenID.QDirection: QDirection = new STFQDirectionItem(subBlock); break;
                        case TokenID.Matrix3x3: Matrix3x3 = new Matrix3x3(subBlock); break;
                        case TokenID.VDbId: VDbId = subBlock.ReadUInt(); break;
                        case TokenID.StaticFlags: StaticFlags = subBlock.ReadFlags(); break;
                        default: subBlock.Skip(); break;
                    }
                }
            }
            // TODO verify that we got all needed parameters otherwise null pointer failures will occur
            // TODO, do this for all objects that iterate using a while loop
        }
    }

    public class StaticObj : BaseObj
    {
        public StaticObj(SBR block, int detailLevel)
            : base(block, detailLevel)
        {
        }
    }

    /// <summary>
    /// Pickup objects supply fuel (diesel, coal) or water.
    /// </summary>
    public class PickupObj : BaseObj
    {
        public readonly uint PickupType;
        public readonly PickupAnimDataItem PickupAnimData;
        public readonly SpeedRangeItem SpeedRange;
        public readonly PickupCapacityItem PickupCapacity;
        public readonly List<TrItemId> TrItemIDList = new List<TrItemId>();
        public readonly uint CollideFlags;

        public WorldLocation Location;

        /// <summary>
        /// Creates the object, but currently skips the animation field.
        /// </summary>
        /// <param name="block"></param>
        /// <param name="detailLevel"></param>
        public PickupObj(SBR block, int detailLevel)
        {
            StaticDetailLevel = detailLevel;

            while (!block.EndOfBlock())
            {
                using (var subBlock = block.ReadSubBlock())
                {
                    switch (subBlock.ID)
                    {
                        case TokenID.UiD: UID = subBlock.ReadUInt(); break;
                        case TokenID.SpeedRange: SpeedRange = new SpeedRangeItem(subBlock); break;
                        case TokenID.PickupType: PickupType = subBlock.ReadUInt();
                            subBlock.Skip(); // Discard the 2nd value (0 or 1 but significance is not known)
                            break;
                        case TokenID.PickupAnimData: PickupAnimData = new PickupAnimDataItem(subBlock); break;
                        case TokenID.PickupCapacity: PickupCapacity = new PickupCapacityItem(subBlock); break;
                        case TokenID.TrItemId: TrItemIDList.Add(new TrItemId(subBlock)); break;
                        case TokenID.CollideFlags: CollideFlags = subBlock.ReadUInt(); break;
                        case TokenID.FileName: FileName = subBlock.ReadString(); break;
                        case TokenID.StaticFlags: StaticFlags = subBlock.ReadFlags(); break;
                        case TokenID.Position: Position = new STFPositionItem(subBlock); break;
                        case TokenID.QDirection: QDirection = new STFQDirectionItem(subBlock); break;
                        case TokenID.VDbId: VDbId = subBlock.ReadUInt(); break;
                        default: subBlock.Skip(); break;
                    }
                }
            }
        }

        /// <summary>
        /// SpeedRangeItem specifies the acceptable range of speeds (meters/sec) for using a pickup.
        /// Presumably non-zero speeds are intended for water troughs or, perhaps, merry-go-round freight.
        /// </summary>
        public class SpeedRangeItem
        {
            public readonly float MinMpS;
            public readonly float MaxMpS;

            internal SpeedRangeItem(SBR block)
            {
                block.VerifyID(TokenID.SpeedRange);
                MinMpS = block.ReadFloat();
                MaxMpS = block.ReadFloat();
                block.VerifyEndOfBlock();
            }
        }

        /// <summary>
        /// PickupAnimDataItem specifies 2 values.  The first represents different pickup animation options.
        /// The second represents the animation speed which will be used.
        /// For the moment PickupOptions may not be used.
        /// </summary>
        public class PickupAnimDataItem
        {
            public readonly float PickupOptions;
            public readonly float AnimationSpeed;

            internal PickupAnimDataItem(SBR block)
            {
                block.VerifyID(TokenID.PickupAnimData);
                PickupOptions = block.ReadFloat();
                AnimationSpeed = block.ReadFloat();
                if (AnimationSpeed == 0) AnimationSpeed = 1.0f;
                block.VerifyEndOfBlock();
            }
        }

        /// <summary>
        /// Creates the object.
        /// The units of measure have been assumed and, once parsed, the values are not currently used.
        /// </summary>
        public class PickupCapacityItem
        {
            public readonly float QuantityAvailableKG;
            public readonly float FeedRateKGpS;

            internal PickupCapacityItem(SBR block)
            {
                block.VerifyID(TokenID.PickupCapacity);
                QuantityAvailableKG = block.ReadFloat();
                FeedRateKGpS = block.ReadFloat();
                block.VerifyEndOfBlock();
            }
        }
    }

    public class TransferObj : WorldObject
    {
        public readonly float Width;
        public readonly float Height;

        public TransferObj(SBR block, int detailLevel)
        {
            StaticDetailLevel = detailLevel;

            while (!block.EndOfBlock())
            {
                using (var subBlock = block.ReadSubBlock())
                {
                    switch (subBlock.ID)
                    {
                        case TokenID.UiD: UID = subBlock.ReadUInt(); break;
                        case TokenID.Width: Width = subBlock.ReadFloat(); break;
                        case TokenID.Height: Height = subBlock.ReadFloat(); break;
                        case TokenID.FileName: FileName = subBlock.ReadString(); break;
                        case TokenID.Position: Position = new STFPositionItem(subBlock); break;
                        case TokenID.QDirection: QDirection = new STFQDirectionItem(subBlock); break;
                        case TokenID.Matrix3x3: Matrix3x3 = new Matrix3x3(subBlock); break;
                        case TokenID.VDbId: VDbId = subBlock.ReadUInt(); break;
                        case TokenID.StaticFlags: StaticFlags = subBlock.ReadFlags(); break;
                        default: subBlock.Skip(); break;
                    }
                }
            }
        }
    }

    public class TrackObj : WorldObject
    {
        public readonly uint SectionIdx;
        public readonly float Elevation;
        public readonly uint CollideFlags;
        public readonly JNodePosn JNodePosn;

        public TrackObj(SBR block, int detailLevel)
        {
            StaticDetailLevel = detailLevel;

            while (!block.EndOfBlock())
            {
                using (var subBlock = block.ReadSubBlock())
                {
                    switch (subBlock.ID)
                    {
                        case TokenID.UiD: UID = subBlock.ReadUInt(); break;
                        case TokenID.SectionIdx: SectionIdx = subBlock.ReadUInt(); break;
                        case TokenID.Elevation: Elevation = subBlock.ReadFloat(); break;
                        case TokenID.CollideFlags: CollideFlags = subBlock.ReadUInt(); break;
                        case TokenID.FileName: FileName = subBlock.ReadString(); break;
                        case TokenID.StaticFlags: StaticFlags = subBlock.ReadUInt(); break;
                        case TokenID.Position: Position = new STFPositionItem(subBlock); break;
                        case TokenID.QDirection: QDirection = new STFQDirectionItem(subBlock); break;
                        case TokenID.Matrix3x3: Matrix3x3 = new Matrix3x3(subBlock); break;
                        case TokenID.VDbId: VDbId = subBlock.ReadUInt(); break;
                        case TokenID.JNodePosn: JNodePosn = new JNodePosn(subBlock); break;
                        default: subBlock.Skip(); break;
                    }
                }
            }
            // TODO verify that we got all needed parameters otherwise null pointer failures will occur
            // TODO, do this for all objects that iterate using a while loop
        }
    }

    public class DyntrackObj : WorldObject
    {
        public readonly uint SectionIdx;
        public readonly float Elevation;
        public readonly uint CollideFlags;
        public readonly TrackSections trackSections;

        public DyntrackObj(SBR block, int detailLevel)
        {
            StaticDetailLevel = detailLevel;

            while (!block.EndOfBlock())
            {
                using (var subBlock = block.ReadSubBlock())
                {
                    switch (subBlock.ID)
                    {
                        case TokenID.UiD: UID = subBlock.ReadUInt(); break;
                        case TokenID.SectionIdx: SectionIdx = subBlock.ReadUInt(); break;
                        case TokenID.Elevation: Elevation = subBlock.ReadFloat(); break;
                        case TokenID.CollideFlags: CollideFlags = subBlock.ReadUInt(); break;
                        case TokenID.StaticFlags: StaticFlags = subBlock.ReadUInt(); break;
                        case TokenID.Position: Position = new STFPositionItem(subBlock); break;
                        case TokenID.QDirection: QDirection = new STFQDirectionItem(subBlock); break;
                        case TokenID.VDbId: VDbId = subBlock.ReadUInt(); break;
                        case TokenID.TrackSections: trackSections = new TrackSections(subBlock); break;
                        default: subBlock.Skip(); break;
                    }
                }
            }
        }

        // DyntrackObj copy constructor with a single TrackSection
        public DyntrackObj(DyntrackObj copy, int iTkSection)
        {
            this.SectionIdx = copy.SectionIdx;
            this.Elevation = copy.Elevation;
            this.CollideFlags = copy.CollideFlags;
            this.StaticFlags = copy.StaticFlags;
            this.Position = new STFPositionItem(copy.Position);
            this.QDirection = new STFQDirectionItem(copy.QDirection);
            this.VDbId = copy.VDbId;
            this.FileName = copy.FileName;
            this.StaticDetailLevel = copy.StaticDetailLevel;
            this.UID = copy.UID;
            //this.totalRealRun = copy.totalRealRun;
            //this.mstsTotalRise = copy.mstsTotalRise;
            // Copy only the single subsection specified
            this.trackSections = new TrackSections(copy.trackSections[iTkSection]);
        }

        public class TrackSections : List<TrackSection>
        {
            public TrackSections()
            {
            }

            public TrackSections(TrackSection trackSection)
            {
                Add(new TrackSection(trackSection));
            }

            public TrackSections(SBR block)
            {
                block.VerifyID(TokenID.TrackSections);
                var count = 5;
                while (count-- > 0)
                    Add(new TrackSection(block.ReadSubBlock(), count));
                block.VerifyEndOfBlock();
            }
        }

        public class TrackSection
        {
            // TrackSection  ==> :SectionCurve :uint,UiD :float,param1 :float,param2
            // SectionCurve  ==> :uint,isCurved
            // eg:  TrackSection (
            //	       SectionCurve ( 1 ) 40002 -0.3 120
            //      )
            // isCurve = 0 for straight, 1 for curved
            // param1 = length (m) for straight, arc (radians) for curved
            // param2 = 0 for straight, radius (m) for curved

            public readonly uint isCurved;
            public readonly uint UiD;
            public readonly float param1;
            public readonly float param2;
            public readonly float deltaY; // Elevation change for this subsection

            public TrackSection(SBR block, int count)
            {
                block.VerifyID(TokenID.TrackSection);
                using (var subBlock = block.ReadSubBlock())
                {
                    subBlock.VerifyID(TokenID.SectionCurve);
                    isCurved = subBlock.ReadUInt();
                    subBlock.VerifyEndOfBlock();
                }
                UiD = block.ReadUInt();
                param1 = block.ReadFloat();
                param2 = block.ReadFloat();
                block.VerifyEndOfBlock();
                deltaY = 0;
            }

            public TrackSection(TrackSection copy)
            {
                this.UiD = copy.UiD;
                this.isCurved = copy.isCurved;
                this.param1 = copy.param1;
                this.param2 = copy.param2;
                this.deltaY = copy.deltaY;
            }
        }
    }

    public class ForestObj : WorldObject
    {
        public readonly bool IsYard;
        public readonly string TreeTexture;
        public readonly int Population;
        public readonly ScaleRange scaleRange;
        public readonly ForestArea forestArea;
        public readonly TreeSize treeSize;

        public ForestObj(SBR block, int detailLevel)
        {
            StaticDetailLevel = detailLevel;

            while (!block.EndOfBlock())
            {
                using (var subBlock = block.ReadSubBlock())
                {
                    switch (subBlock.ID)
                    {
                        case TokenID.UiD: UID = subBlock.ReadUInt(); break;
                        case TokenID.TreeTexture: TreeTexture = subBlock.ReadString(); break;
                        case TokenID.ScaleRange: scaleRange = new ScaleRange(subBlock); break;
                        case TokenID.Area: forestArea = new ForestArea(subBlock); break;
                        case TokenID.Population: Population = subBlock.ReadInt(); break;
                        case TokenID.TreeSize: treeSize = new TreeSize(subBlock); break;
                        case TokenID.StaticFlags: StaticFlags = subBlock.ReadUInt(); break;
                        case TokenID.Position: Position = new STFPositionItem(subBlock); break;
                        case TokenID.QDirection: QDirection = new STFQDirectionItem(subBlock); break;
                        case TokenID.VDbId: VDbId = subBlock.ReadUInt(); break;
                        default: subBlock.Skip(); break;
                    }
                }
            }
            IsYard = TreeTexture == null;
        }

        public class ScaleRange
        {
            public readonly float Minimum;
            public readonly float Maximum;

            internal ScaleRange(SBR block)
            {
                block.VerifyID(TokenID.ScaleRange);
                Minimum = block.ReadFloat();
                Maximum = block.ReadFloat();
                block.VerifyEndOfBlock();
            }
        }

        public class ForestArea
        {
            public readonly float X;
            public readonly float Z;

            internal ForestArea(SBR block)
            {
                block.VerifyID(TokenID.Area);
                X = block.ReadFloat();
                Z = block.ReadFloat();
                block.VerifyEndOfBlock();
            }
        }

        public class TreeSize
        {
            public readonly float Width;
            public readonly float Height;

            internal TreeSize(SBR block)
            {
                block.VerifyID(TokenID.TreeSize);
                Width = block.ReadFloat();
                Height = block.ReadFloat();
                block.VerifyEndOfBlock();
            }
        }
    }

    public class SignalObj : WorldObject
    {
        public readonly uint SignalSubObj;
        public readonly SignalUnits SignalUnits;

        public SignalObj(SBR block, int detailLevel)
        {
            StaticDetailLevel = detailLevel;

            while (!block.EndOfBlock())
            {
                using (var subBlock = block.ReadSubBlock())
                {
                    switch (subBlock.ID)
                    {
                        case TokenID.UiD: UID = subBlock.ReadUInt(); break;
                        case TokenID.FileName: FileName = subBlock.ReadString(); break;
                        case TokenID.Position: Position = new STFPositionItem(subBlock); break;
                        case TokenID.QDirection: QDirection = new STFQDirectionItem(subBlock); break;
                        case TokenID.Matrix3x3: Matrix3x3 = new Matrix3x3(subBlock); break;
                        case TokenID.VDbId: VDbId = subBlock.ReadUInt(); break;
                        case TokenID.StaticFlags: StaticFlags = subBlock.ReadFlags(); break;
                        case TokenID.SignalSubObj: SignalSubObj = subBlock.ReadFlags(); break;
                        case TokenID.SignalUnits: SignalUnits = new SignalUnits(subBlock); break;
                        default: subBlock.Skip(); break;
                    }
                }
            }
        }
    }

    public class SpeedPostObj : WorldObject
    {
        public readonly string Speed_Digit_Tex; //ace
        public readonly Speed_Text_Size Text_Size;// ( 0.08 0.06 0 )
        public readonly Speed_Sign_Shape Sign_Shape;
        public readonly List<TrItemId> trItemIDList = new List<TrItemId>();

        public SpeedPostObj(SBR block, int detailLevel)
        {
            StaticDetailLevel = detailLevel;

            while (!block.EndOfBlock())
            {
                using (var subBlock = block.ReadSubBlock())
                {
                    switch (subBlock.ID)
                    {
                        case TokenID.UiD: UID = subBlock.ReadUInt(); break;
                        case TokenID.Speed_Digit_Tex: Speed_Digit_Tex = subBlock.ReadString(); break;
                        case TokenID.FileName: FileName = subBlock.ReadString(); break;
                        case TokenID.StaticFlags: StaticFlags = subBlock.ReadUInt(); break;
                        case TokenID.Position: Position = new STFPositionItem(subBlock); break;
                        case TokenID.Speed_Sign_Shape: Sign_Shape = new Speed_Sign_Shape(subBlock); break;
                        case TokenID.Speed_Text_Size: Text_Size = new Speed_Text_Size(subBlock); break;
                        case TokenID.QDirection: QDirection = new STFQDirectionItem(subBlock); break;
                        case TokenID.VDbId: VDbId = subBlock.ReadUInt(); break;
                        case TokenID.TrItemId: trItemIDList.Add(new TrItemId(subBlock)); break;
                        default: subBlock.Skip(); break;
                    }
                }
            }
            // TODO verify that we got all needed parameters otherwise null pointer failures will occur
            // TODO, do this for all objects that iterate using a while loop
        }

        public class Speed_Sign_Shape
        {
            public readonly int NumShapes;
            public readonly float[] ShapesInfo; // ( 2 -0.021 0.481 -0.083 0 0 0.475 0.083 3.14159 )

            public Speed_Sign_Shape(SBR block)
            {
                block.VerifyID(TokenID.Speed_Sign_Shape);
                NumShapes = block.ReadInt();
                ShapesInfo = new float[NumShapes * 4];
                for (var i = 0; i < NumShapes; i++)
                {
                    ShapesInfo[i * 4 + 0] = block.ReadFloat();
                    ShapesInfo[i * 4 + 1] = block.ReadFloat();
                    ShapesInfo[i * 4 + 2] = -block.ReadFloat();
                    ShapesInfo[i * 4 + 3] = block.ReadFloat();
                }
                block.VerifyEndOfBlock();
            }
        }

        public class Speed_Text_Size
        {
            public readonly float Size, DX, DY;

            public Speed_Text_Size(SBR block)
            {
                block.VerifyID(TokenID.Speed_Text_Size);
                Size = block.ReadFloat();
                DX = block.ReadFloat();
                DY = block.ReadFloat();
                block.VerifyEndOfBlock();
            }
        }

        public int GetTrItemID(int index)
        {
            var i = 0;
            foreach (var tID in trItemIDList)
            {
                if (tID.db == 0)
                {
                    if (index == i)
                        return tID.dbID;
                    i++;
                }
            }
            return -1;
        }
    }

    public class TrItemId
    {
        public readonly int db, dbID;

        public TrItemId(SBR block)
        {
            block.VerifyID(TokenID.TrItemId);
            db = block.ReadInt();
            dbID = block.ReadInt();
            block.VerifyEndOfBlock();
        }
    }

    public class LevelCrossingObj : WorldObject
    {
        public readonly LevelCrParameters levelCrParameters;
        public readonly LevelCrData levelCrData;
        public readonly LevelCrTiming levelCrTiming;
        public readonly List<TrItemId> trItemIDList = new List<TrItemId>();
        public readonly int crashProbability;
        public readonly bool visible = true;
        public readonly bool silent = false;

        public LevelCrossingObj(SBR block, int detailLevel)
        {
            StaticDetailLevel = detailLevel;

            while (!block.EndOfBlock())
            {
                using (var subBlock = block.ReadSubBlock())
                {
                    switch (subBlock.ID)
                    {
                        case TokenID.UiD: UID = subBlock.ReadUInt(); break;
                        case TokenID.StaticFlags: StaticFlags = subBlock.ReadFlags(); break;
                        case TokenID.LevelCrParameters: levelCrParameters = new LevelCrParameters(subBlock); break;
                        case TokenID.CrashProbability: crashProbability = subBlock.ReadInt(); break;
                        case TokenID.LevelCrData: levelCrData = new LevelCrData(subBlock);
                            visible = (levelCrData.crData1 & 0x1) == 0;
                            silent = !visible || (levelCrData.crData1 & 0x6) == 0x6;
                            break;
                        case TokenID.LevelCrTiming: levelCrTiming = new LevelCrTiming(subBlock); break;
                        case TokenID.TrItemId: trItemIDList.Add(new TrItemId(subBlock)); break;
                        case TokenID.FileName: FileName = subBlock.ReadString(); break;
                        case TokenID.Position: Position = new STFPositionItem(subBlock); break;
                        case TokenID.QDirection: QDirection = new STFQDirectionItem(subBlock); break;
                        case TokenID.VDbId: VDbId = subBlock.ReadUInt(); break;
                        default: subBlock.Skip(); break;
                    }
                }
            }
        }

        public class LevelCrParameters
        {
            public readonly float warningTime, minimumDistance;

            public LevelCrParameters(SBR block)
            {
                block.VerifyID(TokenID.LevelCrParameters);
                warningTime = block.ReadFloat();
                minimumDistance = block.ReadFloat();
                block.VerifyEndOfBlock();
            }
        }

        public class LevelCrData
        {
            public readonly int crData1, crData2; //not known the exact name yet

            public LevelCrData(SBR block)
            {
                block.VerifyID(TokenID.LevelCrData);
                crData1 = block.ReadInt(); // 00000001, should be taken care later, 
                crData2 = block.ReadInt();
                block.VerifyEndOfBlock();
            }
        }

        public class LevelCrTiming
        {
            public readonly float initialTiming, seriousTiming, animTiming;

            public LevelCrTiming(SBR block)
            {
                block.VerifyID(TokenID.LevelCrTiming);
                initialTiming = block.ReadFloat();
                seriousTiming = block.ReadFloat();
                animTiming = block.ReadFloat();
                block.VerifyEndOfBlock();
            }
        }

        public class TrItemId
        {
            public readonly int db, dbID;

            public TrItemId(SBR block)
            {
                block.VerifyID(TokenID.TrItemId);
                db = block.ReadInt();
                dbID = block.ReadInt();
                block.VerifyEndOfBlock();
            }
        }
    }

    public class HazardObj : WorldObject
    {
        public readonly int itemId;

        public HazardObj(SBR block, int detailLevel)
        {
            StaticDetailLevel = detailLevel;

            while (!block.EndOfBlock())
            {
                using (var subBlock = block.ReadSubBlock())
                {
                    switch (subBlock.ID)
                    {
                        case TokenID.UiD: UID = subBlock.ReadUInt(); break;
                        case TokenID.TrItemId: itemId = DecodeTrItemId(subBlock); break;
                        case TokenID.FileName: FileName = subBlock.ReadString(); break;
                        case TokenID.Position: Position = new STFPositionItem(subBlock); break;
                        case TokenID.QDirection: QDirection = new STFQDirectionItem(subBlock); break;
                        case TokenID.VDbId: VDbId = subBlock.ReadUInt(); break;
                        default: subBlock.Skip(); break;
                    }
                }
            }
        }

        int DecodeTrItemId(SBR block)
        {
            block.VerifyID(TokenID.TrItemId);
            int db = block.ReadInt();
            int dbID = block.ReadInt();
            block.VerifyEndOfBlock();
            return dbID;
        }
    }

    public class CarSpawnerObj : WorldObject
    {
        public readonly List<TrItemId> trItemIDList = new List<TrItemId>();
        public readonly float CarFrequency;
        public readonly float CarAvSpeed;

        public CarSpawnerObj(SBR block, int detailLevel)
        {
            StaticDetailLevel = detailLevel;
            CarFrequency = 5.0f;
            CarAvSpeed = 20.0f;

            while (!block.EndOfBlock())
            {
                using (var subBlock = block.ReadSubBlock())
                {
                    switch (subBlock.ID)
                    {
                        case TokenID.UiD: UID = subBlock.ReadUInt(); break;
                        case TokenID.CarFrequency: CarFrequency = subBlock.ReadFloat(); break;
                        case TokenID.CarAvSpeed: CarAvSpeed = subBlock.ReadFloat(); break;
                        case TokenID.TrItemId: trItemIDList.Add(new TrItemId(subBlock)); break;
                        case TokenID.StaticFlags: StaticFlags = subBlock.ReadFlags(); break;
                        case TokenID.Position: Position = new STFPositionItem(subBlock); break;
                        case TokenID.QDirection: QDirection = new STFQDirectionItem(subBlock); break;
                        case TokenID.VDbId: VDbId = subBlock.ReadUInt(); break;
                        default: subBlock.Skip(); break;
                    }
                }
            }
        }

        public int getTrItemID(int index)
        {
            var i = 0;
            foreach (var tID in trItemIDList)
            {
                if (tID.db == 1)
                {
                    if (index == i)
                        return tID.dbID;
                    i++;
                }
            }
            return -1;
        }

        public class TrItemId
        {
            public readonly int db, dbID;

            public TrItemId(SBR block)
            {
                block.VerifyID(TokenID.TrItemId);
                db = block.ReadInt();
                dbID = block.ReadInt();
                block.VerifyEndOfBlock();
            }
        }
    }

    /// <summary>
    /// Super-class for similar track items SidingObj and PlatformObj.
    /// </summary>
    public class TrObject : WorldObject
    {
        public readonly List<TrItemId> trItemIDList = new List<TrItemId>();

        public TrObject(SBR block, int detailLevel)
        {
            StaticDetailLevel = detailLevel;

            while (!block.EndOfBlock())
            {
                using (var subBlock = block.ReadSubBlock())
                {
                    switch (subBlock.ID)
                    {
                        case TokenID.UiD: UID = subBlock.ReadUInt(); break;
                        case TokenID.TrItemId: trItemIDList.Add(new TrItemId(subBlock)); break;
                        case TokenID.StaticFlags: StaticFlags = subBlock.ReadFlags(); break;
                        case TokenID.Position: Position = new STFPositionItem(subBlock); break;
                        case TokenID.QDirection: QDirection = new STFQDirectionItem(subBlock); break;
                        case TokenID.VDbId: VDbId = subBlock.ReadUInt(); break;
                        default: subBlock.Skip(); break;
                    }
                }
            }
        }

        public int getTrItemID(int index)
        {
            var i = 0;
            foreach (var tID in trItemIDList)
            {
                if (tID.db == 0)
                {
                    if (index == i)
                        return tID.dbID;
                    i++;
                }
            }
            return -1;
        }

        public class TrItemId
        {
            public readonly int db, dbID;

            public TrItemId(SBR block)
            {
                block.VerifyID(TokenID.TrItemId);
                db = block.ReadInt();
                dbID = block.ReadInt();
                block.VerifyEndOfBlock();
            }
        }
    }

    /// <summary>
    /// Empty sub-class distinguishes siding objects from platform objects.
    /// </summary>
    public class SidingObj : TrObject
    {
        public SidingObj(SBR block, int detailLevel) :
            base(block, detailLevel)
        {
        }
    }

    /// <summary>
    /// Empty sub-class distinguishes platform objects from siding objects.
    /// </summary>
    public class PlatformObj : TrObject
    {
        public PlatformObj(SBR block, int detailLevel) :
            base(block, detailLevel)
        {
        }
    }

    // These relate to the general properties settable for scenery objects in RE
    public enum StaticFlag
    {
        RoundShadow = 0x00002000,
        RectangularShadow = 0x00004000,
        TreelineShadow = 0x00008000,
        DynamicShadow = 0x00010000,
        AnyShadow = 0x0001E000,
        Terrain = 0x00040000,
        Animate = 0x00080000,
        Global = 0x00200000,
    }

    public abstract class WorldObject
    {
        public string FileName;
        public uint UID;
        public STFPositionItem Position;
        public STFQDirectionItem QDirection;
        public Matrix3x3 Matrix3x3;
        public int StaticDetailLevel;
        public uint StaticFlags;
        public uint VDbId;
    }

    public class STFPositionItem : TWorldPosition
    {
        public STFPositionItem(TWorldPosition p)
            : base(p)
        {
        }

        public STFPositionItem(SBR block)
        {
            block.VerifyID(TokenID.Position);
            X = block.ReadFloat();
            Y = block.ReadFloat();
            Z = block.ReadFloat();
            block.VerifyEndOfBlock();
        }
    }

    public class STFQDirectionItem : TWorldDirection
    {
        public STFQDirectionItem(TWorldDirection d)
            : base(d)
        {
        }

        public STFQDirectionItem(SBR block)
        {
            block.VerifyID(TokenID.QDirection);
            A = block.ReadFloat();
            B = block.ReadFloat();
            C = block.ReadFloat();
            D = block.ReadFloat();
            block.VerifyEndOfBlock();
        }
    }

    public class Matrix3x3
    {
        public readonly float AX, AY, AZ, BX, BY, BZ, CX, CY, CZ;

        public Matrix3x3(SBR block)
        {
            block.VerifyID(TokenID.Matrix3x3);
            AX = block.ReadFloat();
            AY = block.ReadFloat();
            AZ = block.ReadFloat();
            BX = block.ReadFloat();
            BY = block.ReadFloat();
            BZ = block.ReadFloat();
            CX = block.ReadFloat();
            CY = block.ReadFloat();
            CZ = block.ReadFloat();
            block.VerifyEndOfBlock();
        }
    }

    public class JNodePosn
    {
        public readonly int TileX, TileZ;
        public readonly float X, Y, Z;

        public JNodePosn(SBR block)
        {
            block.VerifyID(TokenID.JNodePosn);
            TileX = block.ReadInt();
            TileZ = block.ReadInt();
            X = block.ReadFloat();
            Y = block.ReadFloat();
            Z = block.ReadFloat();
            block.VerifyEndOfBlock();
        }
    }

    public class TWorldDirection
    {
        public float A;
        public float B;
        public float C;
        public float D;

        public TWorldDirection(float a, float b, float c, float d) { A = a; B = b; C = c; D = d; }
        public TWorldDirection() { A = 0; B = 0; C = 0; D = 1; }
        public TWorldDirection(TWorldDirection d) { A = d.A; B = d.B; C = d.C; D = d.D; }

        public void SetBearing(float compassRad)
        {
            var slope = GetSlope();
            SetAngles(compassRad, slope);
        }

        public void SetBearing(float dx, float dz)
        {
            var slope = GetSlope();
            var compassRad = MstsUtility.AngleDxDz(dx, dz);
            SetAngles(compassRad, slope);
        }

        public void Rotate(float radians)  // Rotate around world vertical axis - +degrees is clockwise
        // This rotates about the surface normal
        {
            SetBearing(GetBearing() + radians);
        }

        public void Pivot(float radians)	// This rotates about object Y axis
        {
            radians += GetBearing();
            var slope = GetSlope();
            SetAngles(radians, -slope);
        }

        public void SetAngles(float compassRad, float tiltRad)  // + rad is tilted up or rotated east
        // from http://www.euclideanspace.com/maths/geometry/rotations/conversions/eulerToQuaternion/
        /*
         *  w = Math.sqrt(1.0 + C1 * C2 + C1*C3 - S1 * S2 * S3 + C2*C3) / 2
            x = (C2 * S3 + C1 * S3 + S1 * S2 * C3) / (4.0 * w) 
            y = (S1 * C2 + S1 * C3 + C1 * S2 * S3) / (4.0 * w)
            z = (-S1 * S3 + C1 * S2 * C3 + S2) /(4.0 * w) 


            where:

            C1 = cos(heading) 
            C2 = cos(attitude) 
            C3 = cos(bank) 
            S1 = sin(heading) 
            S2 = sin(attitude) 
            S3 = sin(bank)     it seems in MSTS - tilt forward back is bank
				
        Applied in order of heading, attitude then bank 
        */
        {
            var a1 = compassRad;
            var a2 = 0F;
            var a3 = tiltRad;

            var C1 = (float)Math.Cos(a1);
            var S1 = (float)Math.Sin(a1);
            var C2 = (float)Math.Cos(a2);
            var S2 = (float)Math.Sin(a2);
            var C3 = (float)Math.Cos(a3);
            var S3 = (float)Math.Sin(a3);

            var w = (float)Math.Sqrt(1.0 + C1 * C2 + C1 * C3 - S1 * S2 * S3 + C2 * C3) / 2.0f;
            float x, y, z;

            if (Math.Abs(w) < .000005)
            {
                A = 0.0f;
                B = -1.0f;
                C = 0.0f;
                D = 0.0f;
            }
            else
            {
                x = (float)(-(C2 * S3 + C1 * S3 + S1 * S2 * C3) / (4.0 * w));
                y = (float)(-(S1 * C2 + S1 * C3 + C1 * S2 * S3) / (4.0 * w));
                z = (float)(-(-S1 * S3 + C1 * S2 * C3 + S2) / (4.0 * w));

                A = x;
                B = y;
                C = z;
                D = w;
            }
        }

        public void SetSlope(float tiltRad) // +v is tilted up
        {
            var compassAngleRad = MstsUtility.AngleDxDz(DX(), DZ());
            SetAngles(compassAngleRad, tiltRad);
        }

        public void Tilt(float radians)   // Tilt up the specified number of radians
        {
            SetSlope(GetSlope() + radians);
        }

        public void MakeLevel()  // Remove any tilt from the direction.
        {
            SetSlope(0);
        }

        public float DY()
        {
            var x = -A; // imaginary i part of quaternion
            var y = -B; // imaginary j part of quaternion
            var z = -C; // imaginary k part of quaternion
            var w = D; // real part of quaternionfloat 

            //From http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToEuler/
            //p2.x = ( w*w*p1.x + 2*y*w*p1.z - 2*z*w*p1.y + x*x*p1.x + 2*y*x*p1.y + 2*z*x*p1.z - z*z*p1.x - y*y*p1.x );	
            //p2.y = ( 2*x*y*p1.x + y*y*p1.y + 2*z*y*p1.z + 2*w*z*p1.x - z*z*p1.y + w*w*p1.y - 2*x*w*p1.z - x*x*p1.y );	
            //p2.z = ( 2*x*z*p1.x + 2*y*z*p1.y + z*z*p1.z - 2*w*y*p1.x - y*y*p1.z + 2*w*x*p1.y - x*x*p1.z + w*w*p1.z );

            var dy = (2 * z * y - 2 * x * w);
            return dy;
        }

        public float DX()
        {

            // WAS return -2.0*B*D; 

            /* Was
            float x = C;
            float y = A;
            float z = B;
            float w = D;

            return -2.0 * ( x * y + z * w );
            */

            var x = -A;
            var y = -B;
            var z = -C;
            var w = D;

            var dX = (2 * y * w + 2 * z * x);
            return dX;
        }

        public float DZ()
        {
            var x = -A;
            var y = -B;
            var z = -C;
            var w = D;

            return z * z - y * y - x * x + w * w;
        }

        public float GetSlope()   // Return the slope, +radians is tilted up
        {
            // see http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToEuler/

            var qx = -A;
            var qy = -B;
            var qz = -C;
            var qw = D;

            //float heading;
            //float attitude;
            float bank;

            if (Math.Abs(qx * qy + qz * qw - 0.5) < .00001)
            {
                //heading = 2 * Math.Atan2(qx,qw);
                bank = 0;
            }
            else if (Math.Abs(qx * qy + qz * qw + 0.5) < .00001)
            {
                //heading = -2 * Math.Atan2(qx,qw);
                bank = 0;
            }

            //heading = Math.Atan2(2*qy*qw-2*qx*qz , 1 - 2*qy*qy - 2*qz*qz);
            //attitude = Math.Asin(2*qx*qy + 2*qz*qw);
            bank = (float)Math.Atan2(2 * qx * qw - 2 * qy * qz, 1 - 2 * qx * qx - 2 * qz * qz);

            return bank;
        }

        public float GetBearing()   // Return the bearing
        {
            // see http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToEuler/

            var qx = -A;
            var qy = -B;
            var qz = -C;
            var qw = D;

            float heading;
            //float attitude;
            //float bank;

            if (Math.Abs(qx * qy + qz * qw - 0.5) < .00001)
            {
                heading = 2f * (float)Math.Atan2(qx, qw);
                //bank = 0;
            }
            else if (Math.Abs(qx * qy + qz * qw + 0.5) < .00001)
            {
                heading = -2f * (float)Math.Atan2(qx, qw);
                //bank = 0;
            }
            else
            {
                heading = (float)Math.Atan2(2 * qy * qw - 2 * qx * qz, 1 - 2 * qy * qy - 2 * qz * qz);
                //attitude = Math.Asin(2*qx*qy + 2*qz*qw);
                //bank = Math.Atan2(2*qx*qw-2*qy*qz , 1 - 2*qx*qx - 2*qz*qz);
            }

            return heading;
        }

        public static float AngularDistance(TWorldDirection d1, TWorldDirection d2)
        // number of radians separating angle one and angle two - always positive
        {
            var a1 = d1.GetBearing();
            var a2 = d2.GetBearing();

            var a = a1 - a2;

            a = Math.Abs(a);

            while (a > Math.PI)
                a -= 2.0f * (float)Math.PI;

            return (float)Math.Abs(a);
        }

        /// <summary>
        /// Rotate the specified point in model space to a new location according to the quaternion 
        /// Center of rotation is 0,0,0 in model space
        /// Example   xyz = 0,1,2 rotated 90 degrees east becomes 2,1,0
        /// </summary>
        /// <param name="p1"></param>
        private TWorldPosition RotatePoint(TWorldPosition p1)
        {

            var x = -A; // imaginary i part of quaternion
            var y = -B; // imaginary j part of quaternion
            var z = -C; // imaginary k part of quaternion
            var w = D; // real part of quaternionfloat 

            var p2 = new TWorldPosition();

            p2.X = (w * w * p1.X + 2 * y * w * p1.Z - 2 * z * w * p1.Y + x * x * p1.X + 2 * y * x * p1.Y + 2 * z * x * p1.Z - z * z * p1.X - y * y * p1.X);
            p2.Y = (2 * x * y * p1.X + y * y * p1.Y + 2 * z * y * p1.Z + 2 * w * z * p1.X - z * z * p1.Y + w * w * p1.Y - 2 * x * w * p1.Z - x * x * p1.Y);
            p2.Z = (2 * x * z * p1.X + 2 * y * z * p1.Y + z * z * p1.Z - 2 * w * y * p1.X - y * y * p1.Z + 2 * w * x * p1.Y - x * x * p1.Z + w * w * p1.Z);

            return p2;
        }
    }

    public class TWorldPosition
    {
        public float X;
        public float Y;
        public float Z;

        public TWorldPosition(float x, float y, float z) { X = x; Y = y; Z = z; }
        public TWorldPosition() { X = 0.0f; Y = 0.0f; Z = 0.0f; }
        public TWorldPosition(TWorldPosition p)
        {
            X = p.X;
            Y = p.Y;
            Z = p.Z;
        }

        public void Move(TWorldDirection q, float distance)
        {
            X += (q.DX() * distance);
            Y += (q.DY() * distance);
            Z += (q.DZ() * distance);
        }

        public void Offset(TWorldDirection d, float distanceRight)
        {
            var DRight = new TWorldDirection(d);
            DRight.Rotate(MathHelper.ToRadians(90));
            Move(DRight, distanceRight);
        }

        public static float PointDistance(TWorldPosition p1, TWorldPosition p2)
        // distance between p1 and p2 along the surface
        {
            var dX = p1.X - p2.X;
            var dZ = p1.Z - p2.Z;
            return (float)Math.Sqrt(dX * dX + dZ * dZ);
        }
    }

    public class SignalUnits
    {
        public readonly SignalUnit[] Units;

        public SignalUnits(SBR block)
        {
            var units = new List<SignalUnit>();
            block.VerifyID(TokenID.SignalUnits);
            var count = block.ReadUInt();
            for (var i = 0; i < count; i++)
            {
                using (var subBlock = block.ReadSubBlock())
                {
                    units.Add(new SignalUnit(subBlock));
                }
            }
            block.VerifyEndOfBlock();
            Units = units.ToArray();
        }
    }

    public class SignalUnit
    {
        public readonly int SubObj;
        public readonly uint TrItem;

        public SignalUnit(SBR block)
        {
            block.VerifyID(TokenID.SignalUnit);
            SubObj = block.ReadInt();
            using (var subBlock = block.ReadSubBlock())
            {
                subBlock.VerifyID(TokenID.TrItemId);
                subBlock.ReadUInt(); // Unk?
                TrItem = subBlock.ReadUInt();
                subBlock.VerifyEndOfBlock();
            }
            block.VerifyEndOfBlock();
        }
    }
}
